from django.dispatch import Signal

__all__ = (
    'object_liked',
    'object_unliked'
)


object_liked = Signal(providing_args=["like", "request"])
object_unliked = Signal(providing_args=["object", "request"])



from typing import Tuple

from django.contrib.auth import get_user_model
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.utils.translation import pgettext_lazy

__all__ = (
    'Like',
)

User = get_user_model()


class Like(models.Model):
    """
    Like model

    Attrs:
        sender (ForeignKey): user, which likes an object
        content_type (ForeignKey): content type of a liked object
        object_id (CharField): primary key of a liked object
        content_object (GenericForeignKey): liked object
        created_at (DateTimeField): when object was created

    Methods:
        like (classmethod): implements 'toggle' functionality
    """
    sender = models.ForeignKey(
        User,
        verbose_name=pgettext_lazy("like", "sender"),
        related_name="likes",
        on_delete=models.CASCADE
    )

    content_type = models.ForeignKey(
        ContentType,
        on_delete=models.CASCADE,
        verbose_name=pgettext_lazy('like', 'Content type')
    )
    object_id = models.CharField(
        pgettext_lazy('like', 'Object id'),
        max_length=50
    )
    content_object = GenericForeignKey('content_type', 'object_id')

    created_at = models.DateTimeField(
        verbose_name=pgettext_lazy("like", 'Created at'),
        auto_now_add=True
    )

    class Meta:
        ordering = ['-created_at']
        unique_together = (
            (
                "sender",
                "content_type",
                "object_id"
            ),
        )
        verbose_name = pgettext_lazy('like', 'Like')
        verbose_name_plural = pgettext_lazy('like', 'Likes')

    def __str__(self) -> str:
        return f"{self.sender} - {self.content_object}"

    @classmethod
    def like(
            cls,
            sender: User,
            content_type: ContentType,
            object_id: str
    ) -> Tuple[object, bool]:
        """
        Class method to like-dislike object
        """
        obj, created = cls.objects.get_or_create(
            sender=sender,
            content_type=content_type,
            object_id=object_id
        )
        if not created:
            obj.delete()
        return obj, created



from django.utils.module_loading import import_string

from likes.settings import LIKES_REST_PAGINATION_CLASS

__all__ = (
    'get_pagination_class',
)


def get_pagination_class():
    """
    Returns custom pagination class, set in settings
    """
    pagination_class = LIKES_REST_PAGINATION_CLASS
    if pagination_class:
        try:
            return import_string(pagination_class)
        except ImportError:
            pass



from django.conf import settings as django_setting

__all__ = (
    'LIKES_MODELS',
    'LIKES_REST_PAGINATION_CLASS'
)


LIKES_MODELS = getattr(django_setting, 'LIKES_MODELS', {})
LIKES_REST_PAGINATION_CLASS = getattr(
    django_setting,
    'LIKES_REST_PAGINATION_CLASS',
    None
)



from rest_framework import status
from rest_framework import filters
from rest_framework.generics import ListAPIView, CreateAPIView, GenericAPIView
from rest_framework.response import Response
from rest_framework.views import APIView
from rest_framework.permissions import IsAuthenticated

from likes.api.pagination import get_pagination_class
from likes.api.serializers import (
    LikeSerializer,
    LikeToggleSerializer,
    IsLikedSerializer
)
from likes.models import Like
from likes.services import user_likes_count

__all__ = (
    'LikeListAPIView',
    'LikeToggleView',
    'UserCountOfLikesAPIView',
    'IsLikedAPIView'
)


class LikeListAPIView(ListAPIView):
    """
    List API View to return all likes for authenticated user.
    """
    pagination_class = get_pagination_class()
    permission_classes = (IsAuthenticated, )
    serializer_class = LikeSerializer
    queryset = Like.objects.all()
    filter_backends = (filters.SearchFilter, )
    search_fields = ('content_type__model', )

    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .filter(
                sender=self.request.user
            )
            .select_related('sender')
            .distinct()
        )


class LikeToggleView(CreateAPIView):
    """
    post:
    API View to like-unlike given object by authenticated user.\n
    Possible payload:\n
        {
            "content_type": 1,  // content type id of object
            "id": 1  // object's primary key
        }
    """
    permission_classes = (IsAuthenticated, )
    serializer_class = LikeToggleSerializer

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        self.perform_create(serializer)
        data = serializer.data
        data['is_liked'] = bool(serializer.instance.pk)
        return Response(
            data,
            status=status.HTTP_201_CREATED,
            headers=self.get_success_headers(serializer.data)
        )


class UserCountOfLikesAPIView(APIView):
    """
    API View to return count of likes for authenticated user.
    """
    permission_classes = (IsAuthenticated, )

    def get(self, request, *args, **kwargs):
        return Response(
            data={'count': user_likes_count(request.user)}
        )


class IsLikedAPIView(GenericAPIView):
    """
    post:
    API View to check is given elements are liked by authenticated user.\n
    Possible payload:\n
        {
            "content_type": 1,  // content type id of object
            "ids": [1,2,3]  // list of objects primary keys
        }
    """
    permission_classes = (IsAuthenticated, )
    serializer_class = IsLikedSerializer

    def post(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        if serializer.is_valid():
            return Response(serializer.data)
        return Response(
            data={'errors': serializer.errors},
            status=status.HTTP_400_BAD_REQUEST
        )



from django.contrib.auth import get_user_model
from django.contrib.contenttypes.models import ContentType
from django.http import HttpRequest

from likes.models import Like
from likes.signals import object_liked, object_unliked

__all__ = (
    'user_likes_count',
    'obj_likes_count',
    'is_liked',
    'get_who_liked',
    'send_signals'
)

User = get_user_model()


def user_likes_count(user: User) -> int:
    """
    Returns count of likes for a given user.
    """
    if not user.is_authenticated:
        return 0
    return (
        Like.objects
        .filter(
            sender=user,
            content_type__isnull=False,
            object_id__isnull=False
        )
        .count()
    )


def obj_likes_count(obj) -> int:
    """
    Returns count of likes for a given object.
    """
    return (
        Like.objects
        .filter(
            content_type=ContentType.objects.get_for_model(obj),
            object_id=obj.pk
        )
        .count()
    )


def is_liked(obj, user: User) -> bool:
    """
    Checks if a given object is liked by a given user.
    """
    if not user.is_authenticated:
        return False
    ct = ContentType.objects.get_for_model(obj)
    return (
        Like.objects
        .filter(
            content_type=ct,
            object_id=obj.pk,
            sender=user
        )
        .exists()
    )


def get_who_liked(obj):
    """
    Returns users, who liked a given object.
    """
    ct = ContentType.objects.get_for_model(obj)
    return (
        User.objects
        .filter(
            likes__content_type=ct,
            likes__object_id=obj.pk
        )
    )


def send_signals(
        created: bool,
        request: HttpRequest,
        like: Like,
        obj
):
    """
    Sends signals when object was liked and unliked.
    """
    if created:
        object_liked.send(
            sender=Like,
            like=like,
            request=request
        )
    else:
        object_unliked.send(
            sender=Like,
            object=obj,
            request=request
        )



from django.contrib.auth import get_user_model
from django.contrib.contenttypes.models import ContentType
from django.test import TestCase

from ..models import Like

User = get_user_model()


class LikeTestCase(TestCase):
    def setUp(self):
        self.user_1 = User.objects.create(username="user 1")
        self.user_2 = User.objects.create(username="user 2")
        self.content_type = ContentType.objects.get(model="user")

    def tearDown(self):
        self.user_1.delete()
        self.user_2.delete()

    def test_str(self):
        like = Like(
            sender=self.user_1,
            content_type=self.content_type,
            object_id=self.user_2.pk
        )
        self.assertEquals(str(like), f"{self.user_1.__str__()} - {self.user_2.__str__()}")

    def test_like(self):
        like, liked = Like.like(self.user_1, self.content_type, self.user_2.pk)
        self.assertTrue(liked)
        self.assertIsNotNone(like)
        self.assertEqual(like.sender, self.user_1)

    def test_unlike(self):
        Like.like(self.user_1, self.content_type, self.user_2.pk)
        like, liked = Like.like(self.user_1, self.content_type, self.user_2.pk)
        self.assertFalse(liked)



from django.contrib import admin
from django.utils.html import mark_safe
from django.utils.translation import pgettext_lazy

from likes.models import Like
from likes.utils import admin_change_url

__all__ = (
    'LikeAdmin',
)


@admin.register(Like)
class LikeAdmin(admin.ModelAdmin):
    list_filter = ("created_at", )
    list_display = (
        "sender",
        "content_object_link",
        "created_at"
    )
    list_select_related = ("sender", )
    raw_id_fields = ('sender', )
    readonly_fields = ('created_at', )
    search_fields = (
        "sender__username",
        "sender__email"
    )

    def content_object_link(self, obj):
        if obj.content_object:
            url = admin_change_url(obj.content_object)
            return mark_safe(
                f'<a href="{url}" target="_blank">{obj.content_object}</a>'
            )
        return '-'
    content_object_link.short_description = (
        pgettext_lazy('like', 'Content object')
    )



from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Like',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('object_id', models.CharField(max_length=50, verbose_name='Object id')),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='Created at')),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType', verbose_name='Content type')),
                ('sender', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='likes', to=settings.AUTH_USER_MODEL, verbose_name='sender')),
            ],
            options={
                'verbose_name': 'Like',
                'verbose_name_plural': 'Likes',
                'ordering': ['-created_at'],
            },
        ),
        migrations.AlterUniqueTogether(
            name='like',
            unique_together={('sender', 'content_type', 'object_id')},
        ),
    ]



from django.urls import include, path
from likes.api.views import (
    LikeListAPIView,
    UserCountOfLikesAPIView,
    LikeToggleView,
    IsLikedAPIView
)

app_name = 'likes-api'

urlpatterns = [
    path('likes/', include([
        path('list/', LikeListAPIView.as_view(), name='list'),
        path('count/', UserCountOfLikesAPIView.as_view(), name='count'),
        path('toggle/', LikeToggleView.as_view(), name='toggle'),
        path('is/', IsLikedAPIView.as_view(), name='is'),
    ]))
]



from django.db.models import QuerySet
from django.template import Library

from ..models import Like
from likes.services import (
    obj_likes_count,
    is_liked as is_liked_util,
    get_who_liked
)

register = Library()

__all__ = (
    'likes_count',
    'who_liked',
    'likes',
    'is_liked'
)


@register.simple_tag
@register.filter
def likes_count(obj) -> int:
    """
    Returns count of likes for a given object

    Usage:
        {% likes_count obj %}
    or
        {% likes_count obj as var %}
    or
        {{ obj|likes_count }}
    """
    return obj_likes_count(obj)


@register.simple_tag
def who_liked(obj) -> QuerySet:
    """
    Returns users, who liked a given object.

    Usage:
        {% who_liked object as fans %}
    """
    return get_who_liked(obj)


@register.simple_tag
def likes(user) -> QuerySet:
    """
    Returns likes for a given user

    Usage:
        {% likes request.user as var %}
    """
    return Like.objects.filter(sender=user)


@register.simple_tag
def is_liked(obj, user) -> bool:
    """
    Checks if a given object liked by a given user.

    Usage:
        {% is_liked object request.user as var %}
    """
    return is_liked_util(obj, user)



import json

from django.contrib.auth import get_user_model
from django.contrib.contenttypes.models import ContentType
from django.urls import reverse

from rest_framework import status
from rest_framework.test import APITestCase

from ..models import Like

User = get_user_model()


class BaseAPILikeTestCase(APITestCase):
    def setUp(self):
        self.content_type = ContentType.objects.get(model="user")
        self.username = 'john'
        self.password = 'secret123'
        self.user = User.objects.create_user(
            username=self.username,
            password=self.password,
            email='john@doe.com'
        )
        self.test_user = User.objects.create(username='test')
        Like.like(self.user, self.content_type, self.test_user.pk)

    def tearDown(self):
        self.user.delete()
        self.test_user.delete()


class LikeListAPIViewTestCase(BaseAPILikeTestCase):
    def setUp(self):
        super().setUp()
        self.url = reverse('likes-api:list')

    def test_list(self):
        self.client.login(username=self.username, password=self.password)
        response = self.client.get(self.url)
        self.assertEqual(response.status_code, status.HTTP_200_OK)


class UserCountOfLikesAPIViewTestCase(BaseAPILikeTestCase):
    def setUp(self):
        super().setUp()
        self.url = reverse('likes-api:count')

    def test_count(self):
        self.client.login(username=self.username, password=self.password)
        response = self.client.get(self.url)
        self.assertEqual(response.data.get('count'), 1)


class LikeToggleViewTestCase(BaseAPILikeTestCase):
    def setUp(self):
        super().setUp()
        self.url = reverse('likes-api:toggle')
        self.valid_payload = {
            'id': self.test_user.pk,
            'content_type': self.content_type.pk,
        }

    def test_toggle(self):
        self.client.login(username=self.username, password=self.password)

        self.assertEqual(Like.objects.filter(sender=self.user).count(), 1)

        # test unlike
        with self.settings(LIKES_MODELS={'auth.User': {}}):
            self.client.post(
                self.url,
                data=json.dumps(self.valid_payload),
                content_type='application/json'
            )
            self.assertEqual(Like.objects.filter(sender=self.user).count(), 0)

            # test like
            self.client.post(
                self.url,
                data=json.dumps(self.valid_payload),
                content_type='application/json'
            )
            self.assertEqual(Like.objects.filter(sender=self.user).count(), 1)


class IsLikedAPIViewTestCase(BaseAPILikeTestCase):
    def setUp(self):
        super().setUp()
        self.url = reverse('likes-api:is')
        self.valid_payload = {
            'ids': [self.test_user.pk],
            'content_type': self.content_type.pk,
        }

    def test_is_liked(self):
        self.client.login(username=self.username, password=self.password)
        response = self.client.post(
            self.url,
            data=json.dumps(self.valid_payload),
            content_type='application/json'
        )
        self.assertCountEqual(response.data.get('ids'), [self.test_user.pk])



from django.contrib.auth import get_user_model
from django.contrib.contenttypes.models import ContentType
from django.test import TestCase

from ..models import Like
from ..templatetags.ok_likes import who_liked, likes, likes_count, is_liked

User = get_user_model()


class TemplateTagsTestCase(TestCase):

    def setUp(self):
        self.user = User.objects.create(username="user")
        self.users = []
        for i in range(6):
            self.users.append(
                User.objects.create(username=str(i))
            )
        Like.objects.create(
            sender=self.user,
            content_type=ContentType.objects.get(model="user"),
            object_id=self.users[0].pk
        )

    def test_likes_count(self):
        self.assertEquals(likes_count(self.users[0]), 1)

    def test_is_liked(self):
        self.assertEquals(is_liked(self.users[0], self.user), True)

    def test_likes_all(self):
        qs = likes(self.user)
        self.assertEquals(qs.count(), 1)
        self.assertEquals(qs[0].sender, self.user)
        self.assertEqual(qs[0].content_object, self.users[0])

    def test_who_liked(self):
        fans = list(who_liked(self.users[0]).values_list('id', flat=True))
        self.assertEqual(fans, list(User.objects.filter(pk=self.user.pk).values_list('id', flat=True)))


from django.contrib.auth import get_user_model
from django.contrib.contenttypes.models import ContentType
from django.urls import reverse

from likes.settings import LIKES_MODELS

__all__ = (
    'allowed_content_type',
    'admin_change_url'
)

User = get_user_model()


def allowed_content_type(ct: ContentType) -> bool:
    name = f'{ct.app_label}.{ct.model.title()}'
    if name in LIKES_MODELS.keys():
        return True
    return False


def admin_change_url(obj) -> str:
    """
    Returns admin change url for a given object.
    """
    app_label = obj._meta.app_label
    model_name = obj._meta.model.__name__.lower()
    return reverse(
        f'admin:{app_label}_{model_name}_change',
        args=(obj.pk,)
    )



from django.apps import apps
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ObjectDoesNotExist
from django.utils.module_loading import import_string
from django.utils.translation import pgettext_lazy

from rest_framework import serializers

from likes.models import Like
from likes.settings import LIKES_MODELS
from likes.utils import allowed_content_type
from likes.services import send_signals

__all__ = (
    'LikedObjectRelatedField',
    'LikeSerializer',
    'LikeToggleSerializer',
    'IsLikedSerializer'
)


class LikedObjectRelatedField(serializers.RelatedField):
    """
    A custom field to use for the `liked_object` generic relationship.
    """

    def to_representation(self, value):
        serializers = {}
        for key, val in LIKES_MODELS.items():
            app_label, model_name = key.split('.')
            serializer_path = val.get('serializer')
            if all([app_label, model_name, serializer_path]):
                serializers[
                    apps.get_model(app_label, model_name)
                ] = import_string(serializer_path)
        for model in serializers.keys():
            if isinstance(value, model):
                return serializers[model](
                    instance=value,
                    context=self.context
                ).data
        return str(value)


class LikeSerializer(serializers.ModelSerializer):
    """
    Serializer for Like model
    """
    content_object = LikedObjectRelatedField(read_only=True)

    class Meta:
        model = Like
        fields = '__all__'


class LikeToggleSerializer(serializers.ModelSerializer):
    """
    Serializer to like element
    """
    id = serializers.IntegerField(write_only=True)
    content_type = serializers.PrimaryKeyRelatedField(
        write_only=True, queryset=ContentType.objects.all())

    class Meta:
        model = Like
        fields = ['id', 'content_type']

    def validate_content_type(self, value):
        if not allowed_content_type(value):
            raise serializers.ValidationError(
                pgettext_lazy('like', 'Not allowed content type')
            )
        return value

    def validate(self, data):
        content_type = data['content_type']
        try:
            obj = content_type.get_object_for_this_type(pk=data['id'])
        except ObjectDoesNotExist:
            raise serializers.ValidationError(
                pgettext_lazy('like', 'Object not found.')
            )
        else:
            data['object'] = obj
        return data

    def create(self, validated_data):
        like, created = Like.like(
            self.context['request'].user,
            validated_data['content_type'],
            validated_data['object'].pk
        )
        send_signals(
            created=created,
            request=self.context['request'],
            like=like,
            obj=validated_data['object']
        )
        return like


class IsLikedSerializer(serializers.Serializer):
    """
    Serializer to return liked objects
    """
    ids = serializers.ListField(child=serializers.IntegerField())
    content_type = serializers.PrimaryKeyRelatedField(
        write_only=True, queryset=ContentType.objects.all())

    def validate(self, data):
        liked_ids = []
        for id_ in data['ids']:
            is_liked = (
                Like.objects
                .filter(
                    content_type=data['content_type'],
                    object_id=id_,
                    sender=self.context['request'].user
                ).exists()
            )
            if is_liked:
                liked_ids.append(id_)
        data['ids'] = liked_ids
        return data



from django.apps import apps


if apps.is_installed('django_jinja'):

    from django.db.models import QuerySet

    from django_jinja import library

    from ..models import Like
    from likes.services import (
        obj_likes_count,
        is_liked,
        get_who_liked as who_liked
    )

    __all__ = (
        'get_likes_count',
        'get_who_liked',
        'get_likes',
        'get_is_liked'
    )

    @library.global_function
    def get_likes_count(obj) -> int:
        """
        Returns count of likes for a given object

        Usage:
            {{ get_likes_count(object) }}
        """
        return obj_likes_count(obj)


    @library.global_function
    def get_who_liked(obj) -> QuerySet:
        """
        Returns users, who liked a given object.

        Usage:
            {{ get_who_liked(object) }}
        """
        return who_liked(obj)

    @library.global_function
    def get_likes(user) -> QuerySet:
        """
        Returns likes for a given user

        Usage:
            {{ get_likes(request.user) }}
        """
        return Like.objects.filter(sender=user)

    @library.global_function
    def get_is_liked(obj, user) -> bool:
        """
        Checks if a given object liked by a given user.

        Usage:
            {{ get_is_liked(object, request.user) }}
        """
        return is_liked(obj, user)



from setuptools import setup, find_packages

pkj_name = 'likes'

with open('requirements.txt') as f:
    requires = f.read().splitlines()


setup(
    name='django-ok-likes',
    version='0.4',
    description='Django likes app',
    long_description=open('README.rst').read(),
    author='Oleg Kleschunov',
    author_email='igorkleschunov@gmail.com',
    url='https://github.com/LowerDeez/ok-likes',
    packages=[pkj_name] + [pkj_name + '.' + x for x in find_packages(pkj_name)],
    include_package_data=True,
    license='MIT',
    install_requires=requires,
    classifiers=[
        'Environment :: Web Environment',
        "Development Status :: 4 - Beta",
        "Operating System :: OS Independent",
        "Framework :: Django :: 2.0",
        "Framework :: Django :: 2.1",
        'Intended Audience :: Developers',
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ]

)
